@@ -1,5 +1,4 @@
 #include <endian.h>
-#include <inttypes.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
@@ -64,14 +63,15 @@ void d_raw(unsigned char *buf, unsigned len)
 static void format(char *formatter, size_t fmt_sz, char *tofmt, size_t tofmtsz)
 {
 
-	snprintf(formatter,fmt_sz, "%-*.*s",
+	fmt_sz = snprintf(formatter,fmt_sz, "%-*.*s",
 		 (int)tofmtsz, (int)tofmtsz, tofmt);
 	/* trim() the obnoxious trailing white lines */
-	while (--fmt_sz) {
+	while (fmt_sz) {
 		if (formatter[fmt_sz - 1] != ' ' && formatter[fmt_sz - 1] != '\0') {
 			formatter[fmt_sz] = '\0';
 			break;
 		}
+		fmt_sz--;
 	}
 }
 
@@ -463,8 +463,8 @@ static void show_nvme_id_ns_fpi(__u8 fpi)
 void show_nvme_id_ns(struct nvme_id_ns *ns, unsigned int mode)
 {
 	int i;
-	int human = mode&HUMAN,
-		vs = mode&VS;
+	int human = mode & HUMAN,
+		vs = mode & VS;
 
 	printf("nsze    : %#"PRIx64"\n", (uint64_t)le64_to_cpu(ns->nsze));
 	printf("ncap    : %#"PRIx64"\n", (uint64_t)le64_to_cpu(ns->ncap));
@@ -757,10 +757,10 @@ void show_fw_log(struct nvme_firmware_log_page *fw_log, const char *devname)
 						fw_to_string(fw_log->frs[i]));
 }
 
-unsigned long int48_to_long(__u8 *data)
+uint64_t int48_to_long(__u8 *data)
 {
 	int i;
-	long result = 0;
+	uint64_t result = 0;
 
 	for (i = 0; i < 6; i++) {
 		result *= 256;
@@ -773,12 +773,12 @@ void show_smart_log(struct nvme_smart_log *smart, unsigned int nsid, const char
 {
 	/* convert temperature from Kelvin to Celsius */
 	int c;
-	unsigned int temperature = ((smart->temperature[1] << 8) |
+	int temperature = ((smart->temperature[1] << 8) |
 		smart->temperature[0]) - 273;
 
 	printf("Smart Log for NVME device:%s namespace-id:%x\n", devname, nsid);
 	printf("critical_warning                    : %#x\n", smart->critical_warning);
-	printf("temperature                         : %u C\n", temperature);
+	printf("temperature                         : %d C\n", temperature);
 	printf("available_spare                     : %u%%\n", smart->avail_spare);
 	printf("available_spare_threshold           : %u%%\n", smart->spare_thresh);
 	printf("percentage_used                     : %u%%\n", smart->percent_used);
@@ -815,10 +815,10 @@ void show_intel_smart_log(struct nvme_additional_smart_log *smart, unsigned int
 {
 	printf("Additional Smart Log for NVME device:%s namespace-id:%x\n", devname, nsid);
 	printf("key                               normalized raw\n");
-	printf("program_fail_count              : %3d%%       %lu\n",
+	printf("program_fail_count              : %3d%%       %"PRIu64"\n",
 		smart->program_fail_cnt.norm,
 		int48_to_long(smart->program_fail_cnt.raw));
-	printf("erase_fail_count                : %3d%%       %lu\n",
+	printf("erase_fail_count                : %3d%%       %"PRIu64"\n",
 		smart->erase_fail_cnt.norm,
 		int48_to_long(smart->erase_fail_cnt.raw));
 	printf("wear_leveling                   : %3d%%       min: %u, max: %u, avg: %u\n",
@@ -826,35 +826,35 @@ void show_intel_smart_log(struct nvme_additional_smart_log *smart, unsigned int
 		le16_to_cpu(smart->wear_leveling_cnt.wear_level.min),
 		le16_to_cpu(smart->wear_leveling_cnt.wear_level.max),
 		le16_to_cpu(smart->wear_leveling_cnt.wear_level.avg));
-	printf("end_to_end_error_detection_count: %3d%%       %lu\n",
+	printf("end_to_end_error_detection_count: %3d%%       %"PRIu64"\n",
 		smart->e2e_err_cnt.norm,
 		int48_to_long(smart->e2e_err_cnt.raw));
-	printf("crc_error_count                 : %3d%%       %lu\n",
+	printf("crc_error_count                 : %3d%%       %"PRIu64"\n",
 		smart->crc_err_cnt.norm,
 		int48_to_long(smart->crc_err_cnt.raw));
 	printf("timed_workload_media_wear       : %3d%%       %.3f%%\n",
 		smart->timed_workload_media_wear.norm,
 		((float)int48_to_long(smart->timed_workload_media_wear.raw)) / 1024);
-	printf("timed_workload_host_reads       : %3d%%       %lu%%\n",
+	printf("timed_workload_host_reads       : %3d%%       %"PRIu64"%%\n",
 		smart->timed_workload_host_reads.norm,
 		int48_to_long(smart->timed_workload_host_reads.raw));
-	printf("timed_workload_timer            : %3d%%       %lu min\n",
+	printf("timed_workload_timer            : %3d%%       %"PRIu64" min\n",
 		smart->timed_workload_timer.norm,
 		int48_to_long(smart->timed_workload_timer.raw));
 	printf("thermal_throttle_status         : %3d%%       %u%%, cnt: %u\n",
 		smart->thermal_throttle_status.norm,
 		smart->thermal_throttle_status.thermal_throttle.pct,
 		smart->thermal_throttle_status.thermal_throttle.count);
-	printf("retry_buffer_overflow_count     : %3d%%       %lu\n",
+	printf("retry_buffer_overflow_count     : %3d%%       %"PRIu64"\n",
 		smart->retry_buffer_overflow_cnt.norm,
 		int48_to_long(smart->retry_buffer_overflow_cnt.raw));
-	printf("pll_lock_loss_count             : %3d%%       %lu\n",
+	printf("pll_lock_loss_count             : %3d%%       %"PRIu64"\n",
 		smart->pll_lock_loss_cnt.norm,
 		int48_to_long(smart->pll_lock_loss_cnt.raw));
-	printf("nand_bytes_written              : %3d%%       sectors: %lu\n",
+	printf("nand_bytes_written              : %3d%%       sectors: %"PRIu64"\n",
 		smart->nand_bytes_written.norm,
 		int48_to_long(smart->nand_bytes_written.raw));
-	printf("host_bytes_written              : %3d%%       sectors: %lu\n",
+	printf("host_bytes_written              : %3d%%       sectors: %"PRIu64"\n",
 		smart->host_bytes_written.norm,
 		int48_to_long(smart->host_bytes_written.raw));
 }
@@ -914,6 +914,7 @@ char *nvme_status_to_string(__u32 status)
 	case NVME_SC_LBA_RANGE:			return "LBA_RANGE";
 	case NVME_SC_CAP_EXCEEDED:		return "CAP_EXCEEDED";
 	case NVME_SC_NS_NOT_READY:		return "NS_NOT_READY";
+	case NVME_SC_RESERVATION_CONFLICT:	return "RESERVATION_CONFLICT";
 	case NVME_SC_CQ_INVALID:		return "CQ_INVALID";
 	case NVME_SC_QID_INVALID:		return "QID_INVALID";
 	case NVME_SC_QUEUE_SIZE:		return "QUEUE_SIZE";
@@ -1140,21 +1141,28 @@ void nvme_feature_show_fields(__u32 fid, unsigned int result, unsigned char *buf
 void json_print_list_items(struct list_item *list_items, unsigned len)
 {
 	struct json_object *root;
+	struct json_array *devices;
 	struct json_object *device_attrs;
 	char formatter[41] = { 0 };
 	int index, i = 0;
 	char *product;
 
 	root = json_create_object();
+	devices = json_create_array();
 	for (i = 0; i < len; i++) {
 		device_attrs = json_create_object();
 
 		json_object_add_value_string(device_attrs,
 					     "DevicePath",
 					     list_items[i].node);
+
+		format(formatter, sizeof(formatter),
+			   list_items[i].ctrl.fr,
+			   sizeof(list_items[i].ctrl.fr));
+
 		json_object_add_value_string(device_attrs,
 					     "Firmware",
-					     list_items[i].ctrl.fr);
+					     formatter);
 
 		if (sscanf(list_items[i].node, "/dev/nvme%d", &index) == 1)
 			json_object_add_value_int(device_attrs,
@@ -1171,6 +1179,10 @@ void json_print_list_items(struct list_item *list_items, unsigned len)
 
 		product = nvme_product_name(index);
 
+		json_object_add_value_string(device_attrs,
+					     "ProductName",
+					     product);
+
 		format(formatter, sizeof(formatter),
 		       list_items[i].ctrl.sn,
 		       sizeof(list_items[i].ctrl.sn));
@@ -1179,9 +1191,10 @@ void json_print_list_items(struct list_item *list_items, unsigned len)
 					     "SerialNumber",
 					     formatter);
 
-		json_object_add_value_object(root, product, device_attrs);
+		json_array_add_value_object(devices, device_attrs);
 		free((void*)product);
 	}
+	json_object_add_value_array(root, "Devices", devices);
 	json_print_object(root, NULL);
 }
 
@@ -1422,7 +1435,7 @@ void json_fw_log(struct nvme_firmware_log_page *fw_log, const char *devname)
 
 	for (i = 0; i < 7; i++) {
 		snprintf(fmt, sizeof(fmt), "Firmware Rev Slot %d", i);
-		snprintf(str, sizeof(str), "%lu (%s)", (uint64_t)fw_log->frs[i],
+		snprintf(str, sizeof(str), "%"PRIu64" (%s)", (uint64_t)fw_log->frs[i],
 			 fw_to_string(fw_log->frs[i]));
 		json_object_add_value_string(fwsi, fmt, str);
 	}
@@ -1542,7 +1555,7 @@ void show_registers_cap(struct nvme_bar_cap *cap)
 	printf("\tMaximum Queue Entries Supported (MQES): %u\n\n", cap->mqes + 1);
 }
 
-void show_registers_version(__u32 vs)
+static void show_registers_version(__u32 vs)
 {
 	printf("\tNVMe specification %d.%d\n\n", (vs & 0xffff0000) >> 16,  (vs & 0x0000ff00) >> 8);
 }
@@ -1583,7 +1596,7 @@ static void show_registers_cc_shn (__u8 shn)
 	}
 }
 
-void show_registers_cc (__u32 cc)
+static void show_registers_cc(__u32 cc)
 {
 	printf("\tI/O Completion Queue Entry Size (IOSQES): %u bytes\n", 1 <<   ((cc & 0x00f00000) >> 20));
 	printf("\tI/O Submission Queue Entry Size (IOSQES): %u bytes\n", 1 <<   ((cc & 0x000f0000) >> 16));
@@ -1594,7 +1607,7 @@ void show_registers_cc (__u32 cc)
 	printf("\tEnable                              (EN): %s\n\n", (cc & 0x00000001) ? "Yes":"No");
 }
 
-static void show_registers_csts_shst (__u8 shst)
+static void show_registers_csts_shst(__u8 shst)
 {
 	printf("\tShutdown Status               (SHST): ");
 	switch (shst) {
@@ -1612,7 +1625,7 @@ static void show_registers_csts_shst (__u8 shst)
 	}
 }
 
-void show_registers_csts(__u32 csts)
+static void show_registers_csts(__u32 csts)
 {
 	printf("\tProcessing Paused               (PP): %s\n", (csts & 0x00000020) ? "Yes":"No");
 	printf("\tNVM Subsystem Reset Occurred (NSSRO): %s\n", (csts & 0x00000010) ? "Yes":"No");
@@ -1622,20 +1635,21 @@ void show_registers_csts(__u32 csts)
 
 }
 
-void show_registers_aqa(__u32 aqa)
+static void show_registers_aqa(__u32 aqa)
 {
 	printf("\tAdmin Completion Queue Size (ACQS): %u bytes\n", ((aqa & 0x0fff0000) >> 16)+1);
 	printf("\tAdmin Submission Queue Size (ASQS): %u bytes\n\n", (aqa & 0x00000fff)+1);
 
 }
 
-void show_registers_cmbloc(__u32 cmbloc, __u32 cmbsz)
+static void show_registers_cmbloc(__u32 cmbloc, __u32 cmbsz)
 {
 	if (cmbsz == 0) {
 		printf("\tController Memory Buffer feature is not supported\n\n");
 	}
 	else {
-		printf("\tOffset                 (OFST): %x (See cmbsz.szu for granularity)\n", (cmbloc & 0xfffff000) >> 12);
+		printf("\tOffset                 (OFST): %x (See cmbsz.szu for granularity)\n",
+			(cmbloc & 0xfffff000) >> 12);
 		printf("\tBase Indicator Register (BIR): %x\n\n", cmbloc & 0x00000007 );
 	}
 }
@@ -1654,23 +1668,111 @@ static char *nvme_register_szu_to_string(__u8 szu)
 	}
 }
 
-void show_registers_cmbsz(__u32 cmbsz)
+static void show_registers_cmbsz(__u32 cmbsz)
 {
 	if (cmbsz == 0) {
 		printf("\tController Memory Buffer feature is not supported\n\n");
+		return;
 	}
-	else {
-		printf("\tSize                      (SZ): %u\n", (cmbsz & 0xfffff000) >> 12);
-		printf("\tSize Units               (SZU): %s\n", nvme_register_szu_to_string((cmbsz & 0x00000f00) >> 8));
-		printf("\tWrite Data Support       (WDS): Write Data and metadata transfer in Controller Memory Buffer is %s\n",
-				(cmbsz & 0x00000010) ? "Supported":"Not supported");
-		printf("\tRead Data Support        (RDS): Read Data and metadata transfer in Controller Memory Buffer is %s\n",
-				(cmbsz & 0x00000008) ? "Supported":"Not supported");
-		printf("\tPRP SGL List Support   (LISTS): PRP/SG Lists in Controller Memory Buffer is %s\n",
-				(cmbsz & 0x00000004) ? "Supported":"Not supported");
-		printf("\tCompletion Queue Support (CQS): Admin and I/O Completion Queues in Controller Memory Buffer is %s\n",
-				(cmbsz & 0x00000002) ? "Supported":"Not supported");
-		printf("\tSubmission Queue Support (SQS): Admin and I/O Submission Queues in Controller Memory Buffer is %s\n\n",
-				(cmbsz & 0x00000001) ? "Supported":"Not supported");
+	printf("\tSize                      (SZ): %u\n", (cmbsz & 0xfffff000) >> 12);
+	printf("\tSize Units               (SZU): %s\n", nvme_register_szu_to_string((cmbsz & 0x00000f00) >> 8));
+	printf("\tWrite Data Support       (WDS): Write Data and metadata transfer in Controller Memory Buffer is %s\n",
+			(cmbsz & 0x00000010) ? "Supported":"Not supported");
+	printf("\tRead Data Support        (RDS): Read Data and metadata transfer in Controller Memory Buffer is %s\n",
+			(cmbsz & 0x00000008) ? "Supported":"Not supported");
+	printf("\tPRP SGL List Support   (LISTS): PRP/SG Lists in Controller Memory Buffer is %s\n",
+			(cmbsz & 0x00000004) ? "Supported":"Not supported");
+	printf("\tCompletion Queue Support (CQS): Admin and I/O Completion Queues in Controller Memory Buffer is %s\n",
+			(cmbsz & 0x00000002) ? "Supported":"Not supported");
+	printf("\tSubmission Queue Support (SQS): Admin and I/O Submission Queues in Controller Memory Buffer is %s\n\n",
+			(cmbsz & 0x00000001) ? "Supported":"Not supported");
+}
+
+static inline uint32_t mmio_read32(void *addr)
+{
+	__le32 *p = addr;
+
+	return le32_to_cpu(*p);
+}
+
+/* Access 64-bit registers as 2 32-bit; Some devices fail 64-bit MMIO. */
+static inline __u64 mmio_read64(void *addr)
+{
+	__le32 *p = addr;
+
+	return le32_to_cpu(*p) | ((uint64_t)le32_to_cpu(*(p + 1)) << 32);
+}
+
+void show_ctrl_registers(void *bar, unsigned int mode)
+{
+	uint64_t cap, asq, acq;
+	uint32_t vs, intms, intmc, cc, csts, nssr, aqa, cmbsz, cmbloc;
+
+	int human = mode & HUMAN;
+
+	cap = mmio_read64(bar + NVME_REG_CAP);
+	vs = mmio_read32(bar + NVME_REG_VS);
+	intms = mmio_read32(bar + NVME_REG_INTMS);
+	intmc = mmio_read32(bar + NVME_REG_INTMC);
+	cc = mmio_read32(bar + NVME_REG_CC);
+	csts = mmio_read32(bar + NVME_REG_CSTS);
+	nssr = mmio_read32(bar + NVME_REG_NSSR);
+	aqa = mmio_read32(bar + NVME_REG_AQA);
+	asq = mmio_read64(bar + NVME_REG_ASQ);
+	acq = mmio_read64(bar + NVME_REG_ACQ);
+	cmbloc = mmio_read32(bar + NVME_REG_CMBLOC);
+	cmbsz = mmio_read32(bar + NVME_REG_CMBSZ);
+
+	if (human) {
+		printf("cap     : %"PRIx64"\n", cap);
+		show_registers_cap((struct nvme_bar_cap *)&cap);
+
+		printf("version : %x\n", vs);
+		show_registers_version(vs);
+
+		printf("intms   : %x\n", intms);
+		printf("\tInterrupt Vector Mask Set (IVMS): %x\n\n", intms);
+
+		printf("intmc   : %x\n", intmc);
+		printf("\tInterrupt Vector Mask Clear (IVMC): %x\n\n", intmc);
+
+		printf("cc      : %x\n", cc);
+		show_registers_cc(cc);
+
+		printf("csts    : %x\n", csts);
+		show_registers_csts(csts);
+
+		printf("nssr    : %x\n", nssr);
+		printf("\tNVM Subsystem Reset Control (NSSRC): %u\n\n", nssr);
+
+		printf("aqa     : %x\n", aqa);
+		show_registers_aqa(aqa);
+
+		printf("asq     : %"PRIx64"\n", asq);
+		printf("\tAdmin Submission Queue Base (ASQB): %"PRIx64"\n",
+				asq);
+
+		printf("acq     : %"PRIx64"\n", acq);
+		printf("\tAdmin Completion Queue Base (ACQB): %"PRIx64"\n",
+				acq);
+
+		printf("cmbloc  : %x\n", cmbloc);
+		show_registers_cmbloc(cmbloc, cmbsz);
+
+		printf("cmbsz   : %x\n", cmbsz);
+		show_registers_cmbsz(cmbsz);
+	} else {
+		printf("cap     : %"PRIx64"\n", cap);
+		printf("version : %x\n", vs);
+		printf("intms   : %x\n", intms);
+		printf("intmc   : %x\n", intmc);
+		printf("cc      : %x\n", cc);
+		printf("csts    : %x\n", csts);
+		printf("nssr    : %x\n", nssr);
+		printf("aqa     : %x\n", aqa);
+		printf("asq     : %"PRIx64"\n", asq);
+		printf("acq     : %"PRIx64"\n", acq);
+		printf("cmbloc  : %x\n", cmbloc);
+		printf("cmbsz   : %x\n", cmbsz);
 	}
 }
